from core.rpc import Caller, Server, MessagebusCaller, MessagebusServer
from core.messagebus import RMQConsumer, KafkaConsumer
from component.messagebus import (
    admin, consumer, publisher, message_serializer, mock_consumer
)
from component.log import logger
from config import (
    zrb_app_name, app_broker_type, app_rmq_connection_string,
    app_kafka_bootstrap_servers
)
from ulid import ULID


def create_consumer():
    if app_broker_type == 'rabbitmq':
        return RMQConsumer(
            logger=logger,
            connection_string=app_rmq_connection_string,
            serializer=message_serializer,
            rmq_admin=admin,
            identifier='rmq-rpc-reply-consumer'
        )
    if app_broker_type == 'kafka':
        random_uuid = str(ULID())
        group_id = f'{zrb_app_name}-reply-{random_uuid}'
        return KafkaConsumer(
            logger=logger,
            bootstrap_servers=app_kafka_bootstrap_servers,
            group_id=group_id,
            serializer=message_serializer,
            kafka_admin=admin,
            identifier='kafka-rpc-reply-consumer'
        )
    return mock_consumer


rpc_caller: Caller = MessagebusCaller(
    logger=logger,
    admin=admin,
    publisher=publisher,
    consumer_factory=create_consumer
)

rpc_server: Server = MessagebusServer(
    logger=logger,
    consumer=consumer,
    publisher=publisher
)



from typeguard import typechecked
from typing import Any, Optional
from zrb.helper.string.conversion import to_boolean as conversion_to_boolean
import re
import jinja2

NON_ALPHA_NUM = re.compile(r'[^a-zA-Z0-9]+')


@typechecked
def _is_undefined(value: Any) -> bool:
    return value is None or isinstance(value, jinja2.Undefined)


@typechecked
def coalesce(value: Any, *alternatives: Any) -> Any:
    if not _is_undefined(value):
        return value
    for alternative in alternatives:
        if not _is_undefined(alternative):
            return alternative
    return None


@typechecked
def coalesce_str(value: Any, *alternatives: Any) -> Any:
    if not _is_undefined(value) and value != '':
        return str(value)
    for alternative in alternatives:
        if not _is_undefined(alternative) and alternative != '':
            return str(alternative)
    return ''


@typechecked
def to_camel_case(text: Optional[str]) -> str:
    text = str(text) if not _is_undefined(text) else ''
    pascal = to_pascal_case(text)
    if len(pascal) == 0:
        return pascal
    return pascal[0].lower() + pascal[1:]


@typechecked
def to_pascal_case(text: Optional[str]) -> str:
    text = str(text) if not _is_undefined(text) else ''
    text = _to_alphanum(text)
    return ''.join([
        x.lower().capitalize() for x in _to_space_separated(text).split(' ')
    ])


@typechecked
def to_kebab_case(text: Optional[str]) -> str:
    text = str(text) if not _is_undefined(text) else ''
    text = _to_alphanum(text)
    return '-'.join([
        x.lower() for x in _to_space_separated(text).split(' ')
    ])


@typechecked
def to_snake_case(text: Optional[str]) -> str:
    text = str(text) if not _is_undefined(text) else ''
    text = _to_alphanum(text)
    return '_'.join([
        x.lower() for x in _to_space_separated(text).split(' ')
    ])


@typechecked
def _to_alphanum(text: Optional[str]) -> str:
    text = str(text) if not _is_undefined(text) else ''
    return NON_ALPHA_NUM.sub(' ', text)


@typechecked
def to_human_readable(text: Optional[str]) -> str:
    text = str(text) if not _is_undefined(text) else ''
    return ' '.join([
        x.lower() if x.upper() != x else x
        for x in _to_space_separated(text).split(' ')
    ])


@typechecked
def to_capitalized_human_readable(text: Optional[str]) -> str:
    return to_human_readable(text).capitalize()


@typechecked
def _to_space_separated(text: Optional[str]) -> str:
    text = str(text) if not _is_undefined(text) else ''
    text = text.replace('-', ' ').replace('_', ' ')
    parts = text.split(' ')
    new_parts = []
    for part in parts:
        new_part = ''
        for char_index, char in enumerate(part):
            is_first = char_index == 0
            is_last = char_index == len(part) - 1
            previous_char = part[char_index - 1] if not is_first else ''
            next_char = part[char_index + 1] if not is_last else ''
            if char.isupper() and char != ' ' and (
                (not is_last and next_char.islower()) or
                (not is_first and previous_char.islower())
            ):
                new_part += ' ' + char
                continue
            new_part += char
        new_part = new_part.strip(' ')
        if new_part != '':
            new_parts.append(new_part)
    return ' '.join(new_parts).strip(' ')


@typechecked
def to_boolean(text: str) -> bool:
    return conversion_to_boolean(text)



from pulumi_kubernetes.helm.v3 import Chart, LocalChartOpts
from _common import (
    NAMESPACE, RABBITMQ_AUTH_USERNAME, RABBITMQ_AUTH_PASSWORD
)


def create_rabbitmq() -> Chart:
    rabbitmq = Chart(
        'rabbitmq',
        LocalChartOpts(
            path='./helm-charts/rabbitmq',
            namespace=NAMESPACE,
            values={
                'auth': {
                    'username': RABBITMQ_AUTH_USERNAME,
                    'password': RABBITMQ_AUTH_PASSWORD
                },
                'resources': {
                    'limits': {
                        'cpu': '1000m',
                        'memory': '2Gi'
                    },
                    'requests': {
                        'cpu': '1000m',
                        'memory': '1Gi'
                    },
                }
            },
            skip_await=True
        ),
    )
    return rabbitmq



from zrb import (
    runner, CmdTask, ResourceMaker, DockerComposeTask, FlowTask, FlowNode,
    Env, StrInput, HTTPChecker
)
import os
import tomli

CURRENT_DIR = os.path.dirname(__file__)

with open(os.path.join(CURRENT_DIR, 'pyproject.toml'), 'rb') as f:
    toml_dict = tomli.load(f)
    VERSION = toml_dict['project']['version']

###############################################################################
# Input Definitions
###############################################################################

zrb_image_name_input = StrInput(
    name='zrb-image-name',
    description='Zrb image name',
    prompt='Zrb image name',
    default=f'docker.io/stalchmst/zrb:{VERSION}'
)

zrb_latest_image_name_input = StrInput(
    name='zrb-latest-image-name',
    description='Zrb latest image name',
    prompt='Zrb latest image name',
    default='docker.io/stalchmst/zrb:latest'
)

###############################################################################
# Env Definitions
###############################################################################

zrb_image_env = Env(
    name='ZRB_IMAGE',
    os_name='',
    default='{{input.zrb_image_name}}'
)

zrb_latest_image_env = Env(
    name='ZRB_IMAGE',
    os_name='',
    default='{{input.zrb_latest_image_name}}'
)

###############################################################################
# Task Definitions
###############################################################################

build = CmdTask(
    name='build',
    description='Build Zrb',
    cwd=CURRENT_DIR,
    cmd=[
        'set -e',
        'echo "🤖 Build zrb distribution"',
        f'rm -Rf {CURRENT_DIR}/dist',
        'git add . -A',
        'flit build',
    ],
)
runner.register(build)

publish_pip = CmdTask(
    name='publish-pip',
    description='Publish zrb to pypi',
    upstreams=[build],
    cwd=CURRENT_DIR,
    cmd=[
        'set -e',
        'echo "🤖 Publish zrb to pypi"',
        'flit publish --repository pypi',
    ]
)
runner.register(publish_pip)

publish_pip_test = CmdTask(
    name='publish-pip-test',
    description='Publish zrb to testpypi',
    upstreams=[build],
    cwd=CURRENT_DIR,
    cmd=[
        'set -e',
        'echo "🤖 Publish zrb to testpypi"',
        'flit publish --repository testpypi',
    ]
)
runner.register(publish_pip_test)

prepare_docker = ResourceMaker(
    name='prepare-docker',
    description='Create docker directory',
    template_path=f'{CURRENT_DIR}/docker-template',
    destination_path=f'{CURRENT_DIR}/.docker-dir',
    replacements={
        'ZRB_VERSION': VERSION
    }
)
runner.register(prepare_docker)

check_pip = HTTPChecker(
    name='check-pip',
    is_https=True,
    host='pypi.org',
    url=f'pypi/zrb/{VERSION}/json',
    port=443
)

build_image = DockerComposeTask(
    name='build-image',
    description='Build docker image',
    upstreams=[
        prepare_docker,
        check_pip,
    ],
    inputs=[zrb_image_name_input],
    envs=[zrb_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='build',
    compose_args=['zrb']
)
runner.register(build_image)

build_latest_image = DockerComposeTask(
    name='build-latest-image',
    description='Build docker image',
    upstreams=[
        prepare_docker,
        check_pip,
        build_image,
    ],
    inputs=[zrb_latest_image_name_input],
    envs=[zrb_latest_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='build',
    compose_args=['zrb']
)
runner.register(build_latest_image)

stop_container = DockerComposeTask(
    name='stop-container',
    description='remove docker container',
    upstreams=[prepare_docker],
    inputs=[zrb_image_name_input],
    envs=[zrb_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='down'
)
runner.register(stop_container)

start_container = DockerComposeTask(
    name='start-container',
    description='Run docker container',
    upstreams=[
        build_image,
        stop_container
    ],
    inputs=[zrb_image_name_input],
    envs=[zrb_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='up',
    compose_flags=['-d']
)
runner.register(start_container)

push_image = DockerComposeTask(
    name='push-image',
    description='Push docker image',
    upstreams=[build_image],
    inputs=[zrb_image_name_input],
    envs=[zrb_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='push',
    compose_args=['zrb']
)
runner.register(push_image)

push_latest_image = DockerComposeTask(
    name='push-latest-image',
    description='Push docker image',
    upstreams=[
        build_latest_image,
        push_image,
    ],
    inputs=[zrb_latest_image_name_input],
    envs=[zrb_latest_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='push',
    compose_args=['zrb']
)
runner.register(push_latest_image)

publish = FlowTask(
    name='publish',
    description='Publish new version',
    nodes=[
        FlowNode(task=publish_pip),
        FlowNode(task=push_latest_image),
    ]
)
runner.register(publish)

install_symlink = CmdTask(
    name='install-symlink',
    description='Install Zrb as symlink',
    upstreams=[build],
    cmd=[
        'set -e',
        f'cd {CURRENT_DIR}',
        'echo "🤖 Install zrb"',
        'flit install --symlink',
    ]
)
runner.register(install_symlink)

test = CmdTask(
    name='test',
    description='Run zrb test',
    inputs=[
        StrInput(
            name='test',
            shortcut='t',
            description='Specific test case (i.e., test/file.py::test_name)',
            prompt='Test (i.e., test/file.py::test_name)',
            default=''
        ),
    ],
    upstreams=[install_symlink],
    cmd=[
        'set -e',
        f'cd {CURRENT_DIR}',
        'echo "🤖 Perform test"',
        'pytest --ignore-glob="**/template/**/test" --ignore=playground --cov=zrb --cov-report=html --cov-report=term --cov-report=term-missing {{input.test}}'  # noqa
    ],
    retry=0,
    checking_interval=1
)
runner.register(test)

serve_test = CmdTask(
    name='serve-test',
    description='Serve zrb test result',
    inputs=[
        StrInput(
            name='port',
            shortcut='p',
            description='Port to serve coverage result',
            prompt='Serve coverage on port',
            default='9000'
        )
    ],
    upstreams=[test],
    cmd=[
        'set -e',
        f'cd {CURRENT_DIR}',
        'echo "🤖 Serve coverage report"',
        f'python -m http.server {{input.port}} --directory {CURRENT_DIR}/htmlcov',  # noqa
    ],
    checkers=[
        HTTPChecker(port='{{input.port}}')
    ],
    retry=0,
    checking_interval=0.3
)
runner.register(serve_test)

playground = CmdTask(
    name='playground',
    upstreams=[install_symlink],
    cmd=[
        'set -e',
        f'cd {CURRENT_DIR}',
        'echo "🤖 Remove playground"',
        'sudo rm -Rf playground',
        'echo "🤖 Create playground"',
        'cp -R playground-template playground',
        f'cd {CURRENT_DIR}/playground',
        'echo "🤖 Seed project"',
        './seed-project.sh',
        'echo "🤖 Change to playground directory:"',
        f'echo "      cd {CURRENT_DIR}/playground"',
        'echo "🤖 Or playground project directory:"',
        f'echo "      cd {CURRENT_DIR}/playground/my-project"',
        'echo "🤖 You can also test pip package:"',
        f'echo "      cd {CURRENT_DIR}/playground/my-project/src/zrb-coba-test"',  # noqa
        'echo "      source .venv/bin/activate"',
        'echo "      zrb-coba-test"',
        'echo "🤖 And start hacking around. Good luck :)"',
    ],
    retry=0,
    preexec_fn=None
)
runner.register(playground)



from zrb.builtin.group import project_group
from zrb import Task, runner

push_project_images = Task(
    name='push-images',
    group=project_group,
    upstreams=[],
    description='Build project images',
    run=lambda *args, **kwargs: kwargs.get('_task').print_out('🆗')
)
runner.register(push_project_images)



from typing import Any, Callable, Iterable, List, Optional, TypeVar, Union
from typeguard import typechecked
from zrb.task.base_task import BaseTask
from zrb.task.any_task import AnyTask
from zrb.task.task import Task
from zrb.task.cmd_task import CmdTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput
from zrb.helper.accessories.name import get_random_name

import copy

TFlowNode = TypeVar('TFlowNode', bound='FlowNode')


@typechecked
class FlowNode():
    def __init__(
        self,
        name: str = '',
        task: Optional[AnyTask] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        upstreams: Iterable[AnyTask] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        cmd: Union[str, Iterable[str]] = '',
        cmd_path: str = '',
        preexec_fn: Optional[Callable[[], Any]] = None,
        run: Optional[Callable[..., Any]] = None,
        nodes: List[Union[TFlowNode, List[TFlowNode]]] = []
    ):
        self._name = name if name != '' else get_random_name()
        self._task = task
        self._inputs = inputs
        self._envs = envs
        self._env_files = env_files
        self._upstreams = upstreams
        self._icon = icon
        self._color = color
        self._cmd = cmd
        self._cmd_path = cmd_path
        self._preexec_fn = preexec_fn
        self._run_function = run
        self._nodes = nodes

    def to_task(
        self,
        upstreams: List[AnyTask] = [],
        inputs: List[AnyInput] = [],
        envs: List[Env] = [],
        env_files: List[EnvFile] = [],
    ):
        if self._task is not None:
            task = copy.deepcopy(self._task)
            additional_upstreams = self._upstreams + upstreams
            if len(upstreams) > 0:
                task.add_upstreams(*additional_upstreams)
            return task
        if self._run_function is not None:
            return Task(
                name=self._name,
                upstreams=upstreams + self._upstreams,
                inputs=inputs + self._inputs,
                envs=envs + self._envs,
                env_files=env_files + self._env_files,
                icon=self._icon,
                color=self._color,
                run=self._run_function
            )
        if self._cmd != '' or self._cmd_path != '':
            return CmdTask(
                name=self._name,
                upstreams=upstreams + self._upstreams,
                inputs=inputs + self._inputs,
                envs=envs + self._envs,
                env_files=env_files + self._env_files,
                icon=self._icon,
                color=self._color,
                cmd=self._cmd,
                cmd_path=self._cmd_path,
                preexec_fn=self._preexec_fn
            )
        return FlowTask(
            name=self._name,
            upstreams=upstreams + self._upstreams,
            inputs=inputs + self._inputs,
            envs=envs + self._envs,
            env_files=env_files + self._env_files,
            icon=self._icon,
            color=self._color,
            nodes=self._nodes
        )


@typechecked
class FlowTask(BaseTask):

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: float = 0,
        retry: int = 2,
        retry_interval: float = 1,
        nodes: List[Union[FlowNode, List[FlowNode]]] = [],
        skip_execution: Union[bool, str, Callable[..., bool]] = False
    ):
        final_upstreams: List[AnyTask] = upstreams
        for node in nodes:
            flow_nodes = self._get_flow_nodes(node)
            new_upstreams: List[AnyTask] = [
                flow_node.to_task(
                    upstreams=final_upstreams,
                    inputs=inputs,
                    envs=envs,
                    env_files=env_files
                )
                for flow_node in flow_nodes
            ]
            final_upstreams = new_upstreams
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=final_upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            skip_execution=skip_execution,
            run=lambda *args, **kwargs: kwargs.get('_task').print_out('🆗')
        )

    def _get_flow_nodes(
        self, node: Union[FlowNode, List[FlowNode]]
    ) -> List[FlowNode]:
        if isinstance(node, FlowNode):
            return [node]
        return node

    def __repr__(self) -> str:
        return f'<FlowTask name={self._name}>'



from typing import Any, Callable, Iterable, Optional, Union
from typeguard import typechecked
from http.client import HTTPConnection, HTTPSConnection
from zrb.task.base_task import BaseTask
from zrb.task.any_task import AnyTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

import asyncio


@typechecked
class HTTPChecker(BaseTask):

    def __init__(
        self,
        name: str = 'http-check',
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        host: str = 'localhost',
        port: Union[int, str] = 80,
        timeout: Union[int, str] = 5,
        method: str = 'HEAD',
        url: str = '/',
        is_https: Union[bool, str] = False,
        upstreams: Iterable[AnyTask] = [],
        checking_interval: float = 0.1,
        show_error_interval: float = 5,
        skip_execution: Union[bool, str, Callable[..., bool]] = False
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=[],
            checking_interval=checking_interval,
            retry=0,
            retry_interval=0,
            skip_execution=skip_execution
        )
        self._host = host
        self._port = port
        self._timeout = timeout
        self._method = method
        self._url = url
        self._is_https = is_https
        self._show_error_interval = show_error_interval

    def to_function(
        self, env_prefix: str = '', raise_error: bool = True
    ) -> Callable[..., bool]:
        return super().to_function(env_prefix, raise_error)

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        is_https = self.render_bool(self._is_https)
        method = self.render_str(self._method)
        host = self.render_str(self._host)
        port = self.render_int(self._port)
        url = self.render_str(self._url)
        if not url.startswith('/'):
            url = '/' + url
        timeout = self.render_int(self._timeout)
        wait_time = 0
        while not self._check_connection(
            method=method,
            host=host,
            port=port,
            url=url,
            is_https=is_https,
            timeout=timeout,
            should_print_error=wait_time >= self._show_error_interval
        ):
            if wait_time >= self._show_error_interval:
                wait_time = 0
            await asyncio.sleep(self._checking_interval)
            wait_time += self._checking_interval
        return True

    def _check_connection(
        self,
        method: str,
        host: str,
        port: int,
        url: str,
        is_https: bool,
        timeout: int,
        should_print_error: bool
    ) -> bool:
        label = self._get_label(method, host, port, is_https, url)
        conn = self._get_connection(host, port, is_https, timeout)
        try:
            conn.request(method, url)
            res = conn.getresponse()
            if res.status < 300:
                self.log_info('Connection success')
                self.print_out(f'{label} {res.status} (OK)')
                return True
            self._debug_and_print_error(
                f'{label} {res.status} (Not OK)', should_print_error
            )
        except Exception:
            self._debug_and_print_error(
                f'{label} Connection error', should_print_error
            )
        finally:
            conn.close()
        return False

    def _debug_and_print_error(self, message: str, should_print_error: bool):
        if should_print_error:
            self.print_err(message)
        self.log_debug(message)

    def _get_label(
        self, method: str, host: str, port: int, is_https: bool, url: str
    ) -> str:
        protocol = 'https' if is_https else 'http'
        return f'{method} {protocol}://{host}:{port}{url}'

    def _get_connection(
        self, host: str, port: int, is_https: bool, timeout: int
    ) -> Union[HTTPConnection, HTTPSConnection]:
        if is_https:
            return HTTPSConnection(host, port, timeout=timeout)
        return HTTPConnection(host, port, timeout=timeout)

    def __repr__(self) -> str:
        return f'<HttpChecker name={self._name}>'



from typeguard import typechecked
from zrb.helper.log import logger
from zrb.helper.accessories.color import colored

import importlib.util
import os
import re

pattern = re.compile('[^a-zA-Z0-9]')


@typechecked
def load_module(script_path: str):
    if not os.path.isfile(script_path):
        return
    module_name = pattern.sub('', script_path)
    logger.info(colored('Get module spec', attrs=['dark']))
    spec = importlib.util.spec_from_file_location(module_name, script_path)
    logger.info(colored('Create module', attrs=['dark']))
    module = importlib.util.module_from_spec(spec)
    logger.info(colored('Exec module', attrs=['dark']))
    spec.loader.exec_module(module)
    logger.info(colored('Module executed', attrs=['dark']))



from typing import Any
from typeguard import typechecked
from zrb.builtin.group import explain_group
from zrb.task.decorator import python_task
from zrb.task.task import Task
from zrb.runner import runner


@typechecked
def show_lines(task: Task, *lines: str):
    separator = '\n    '
    task.print_out(
        '\n' + separator + separator.join(lines) + '\n', trim_message=False
    )


###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='solid-principle',
    group=explain_group,
    description='Explain SOLID principle',
    runner=runner
)
async def explain_solid_principle(*args: Any, **kwargs: Any):
    show_lines(
        kwargs['_task'],
        'S - Single Responsibility Principle',
        'O - Open/Closed Principle',
        'L - Liskov’s Substitution Principle',
        'I - Interface Segregation Principle',
        'D - Dependency Inversion Principle',
        '',
        'In software development, Object-Oriented Design plays a crucial role when it comes to writing flexible, scalable, maintainable, and reusable code. There are so many benefits of using OOD but every developer should also have the knowledge of the SOLID principle for good object-oriented design in programming. The SOLID principle was introduced by Robert C. Martin, also known as Uncle Bob and it is a coding standard in programming.', # noqa
        '',
        'The SOLID principle helps in reducing tight coupling. Tight coupling means a group of classes are highly dependent on one another which you should avoid in your code. Opposite of tight coupling is loose coupling and your code is considered as a good code when it has loosely-coupled classes. Loosely coupled classes minimize changes in your code, helps in making code more reusable, maintainable, flexible and stable.', # noqa
    )


@python_task(
    name='yagni-principle',
    group=explain_group,
    description='Explain YAGNI principle',
    runner=runner
)
async def explain_yagni_principle(*args: Any, **kwargs: Any):
    show_lines(
        kwargs['_task'],
        'Y - You',
        'A - Aren\'t',
        'G - Gonna',
        'N - Need',
        'I - It',
        '',
        'YAGNI principle ("You Aren\'t Gonna Need It") is a practice in software development which states that features should only be added when required.', # noqa
        '',
        'As a part of the extreme programming (XP) philosophy, YAGNI trims away excess and inefficiency in development to facilitate the desired increased frequency of releases.', # noqa
    )


@python_task(
    name='dry-principle',
    group=explain_group,
    description='Explain DRY principle',
    runner=runner
)
async def explain_dry_principle(*args: Any, **kwargs: Any):
    show_lines(
        kwargs['_task'],
        'D - Don\'t',
        'R - Repeat',
        'Y - Yourself',
        '',
        '"Don\'t repeat yourself" (DRY) is a principle of software development aimed at reducing repetition of information which is likely to change, replacing it with abstractions that are less likely to change, or using data normalization which avoids redundancy in the first place.', # noqa
        '',
        'The DRY principle is stated as "Every piece of knowledge must have a single, unambiguous, authoritative representation within a system". The principle has been formulated by Andy Hunt and Dave Thomas in their book The Pragmatic Programmer.', # noqa
    )


@python_task(
    name='kiss-principle',
    group=explain_group,
    description='Explain KISS principle',
    runner=runner
)
async def explain_kiss_principle(*args: Any, **kwargs: Any):
    show_lines(
        kwargs['_task'],
        'K - Keep',
        'I - It',
        'S - Simple',
        'S - Stupid',
        '',
        'KISS, an acronym for "Keep it simple, stupid!", is a design principle noted by the U.S. Navy in 1960. First seen partly in American English by at least 1938, the KISS principle states that most systems work best if they are kept simple rather than made complicated; therefore, simplicity should be a key goal in design, and unnecessary complexity should be avoided.' # noqa
    )


@python_task(
    name='zen-of-python',
    group=explain_group,
    description='Explain Zen of Python',
    runner=runner
)
async def explain_zen_of_python(*args: Any, **kwargs: Any):
    show_lines(
        kwargs['_task'],
        'Beautiful is better than ugly.',
        'Explicit is better than implicit.',
        'Simple is better than complex.',
        'Complex is better than complicated.',
        'Flat is better than nested.',
        'Sparse is better than dense.',
        'Readability counts.',
        'Special cases aren\'t special enough to break the rules.',
        'Although practicality beats purity.',
        'Errors should never pass silently.',
        'Unless explicitly silenced.',
        'In the face of ambiguity, refuse the temptation to guess.',
        'There should be one-- and preferably only one --obvious way to do it.', # noqa
        'Although that way may not be obvious at first unless you\'re Dutch.',
        'Now is better than never.',
        'Although never is often better than *right* now.',
        'If the implementation is hard to explain, it\'s a bad idea.',
        'If the implementation is easy to explain, it may be a good idea.',
        'Namespaces are one honking great idea -- let\'s do more of those!',
    )



import os

MESSAGE = os.getenv('APP_MESSAGE', 'Hello, world!')
PORT = int(os.getenv('APP_PORT', '8000'))


async def app(scope, receive, send):
    assert scope['type'] == 'http'

    await send({
        'type': 'http.response.start',
        'status': 200,
        'headers': [
            [b'content-type', b'text/plain'],
        ],
    })
    await send({
        'type': 'http.response.body',
        'body': bytes(MESSAGE, 'utf-8'),
    })



from module.auth.component.password_hasher import password_hasher
from component.log import logger
from component.db_connection import engine
from module.auth.entity.user.repo import (
    UserRepo, UserDBRepo
)

user_repo: UserRepo = UserDBRepo(
    logger=logger, engine=engine, password_hasher=password_hasher
)



from typing import Any, List, Optional
from typeguard import typechecked
from zrb.builtin.generator.common.task_input import project_dir_input
from zrb.builtin.generator.common.task_factory import create_add_upstream
from zrb.task.decorator import python_task
from zrb.task.any_task import AnyTask
from zrb.task.task import Task
from zrb.task_input.any_input import AnyInput
from zrb.helper.codemod.add_import_module import add_import_module
from zrb.helper.codemod.add_assert_resource import add_assert_resource
from zrb.helper.file.copy_tree import copy_tree
from zrb.helper.file.text import read_text_file_async, write_text_file_async
import os

CURRENT_DIR = os.path.dirname(__file__)


@typechecked
def create_ensure_project_tasks(
    upstreams: Optional[List[AnyTask]] = None
) -> Task:
    '''
    Create a task to ensure there are project tasks under `_automate/_project`
    '''
    @python_task(
        name='ensure-project-tasks',
        inputs=[project_dir_input],
        upstreams=[] if upstreams is None else upstreams
    )
    async def _task(*args: Any, **kwargs: Any):
        task: Task = kwargs.get('_task')
        project_dir = kwargs.get('project_dir', '.')
        task.print_out('Create project tasks if not exist')
        if os.path.exists(os.path.join(project_dir, '_automate', '_project')):
            return
        await copy_tree(
            src=os.path.join(CURRENT_DIR, 'template'),
            dst=project_dir
        )
        project_task_module_path = '_automate._project'
        zrb_init_path = os.path.join(project_dir, 'zrb_init.py')
        code = await read_text_file_async(zrb_init_path)
        import_alias = project_task_module_path.split('.')[-1]
        code = add_import_module(
            code=code,
            module_path=project_task_module_path,
            alias=import_alias
        )
        code = add_assert_resource(code, import_alias)
        await write_text_file_async(zrb_init_path, code)
    return _task


@typechecked
def create_add_build_images_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file('build_project_images.py'),
        task_name='build-images',
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams
    )


@typechecked
def create_add_deploy_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file('deploy_project.py'),
        task_name='deploy',
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams
    )


@typechecked
def create_add_destroy_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file('destroy_project.py'),
        task_name='destroy',
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams
    )


@typechecked
def create_add_push_images_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file('push_project_images.py'),
        task_name='push-images',
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams
    )


@typechecked
def create_add_remove_containers_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file('remove_project_containers.py'),
        task_name='remove-containers',
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams
    )


@typechecked
def create_add_start_containers_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file('start_project_containers.py'),
        task_name='start-containers',
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams
    )


@typechecked
def create_add_start_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file('start_project.py'),
        task_name='start',
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams
    )


@typechecked
def create_add_stop_containers_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file('stop_project_containers.py'),
        task_name='stop-containers',
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams
    )


@typechecked
def _get_project_task_file(task_file_name: str) -> str:
    return os.path.join('_automate', '_project', task_file_name)



from typeguard import typechecked
from typing import Any, Mapping


@typechecked
def to_str(
    str_map: Mapping[str, Any],
    item_separator: str = '\n',
    item_prefix: str = '',
    keyval_separator: str = ' : '
) -> str:
    keys = list(str_map.keys())
    keys.sort()
    str_list = [
        item_prefix + key + keyval_separator + str(str_map[key])
        for key in keys
    ]
    return item_separator.join(str_list)



from typing import Any
from zrb.builtin.generator.fastapp_field.helper import (
    add_column_to_delete_page, add_column_to_detail_page,
    add_column_to_insert_page, add_column_to_list_page, add_column_to_repo,
    add_column_to_schema, add_column_to_test, add_column_to_update_page
)
from zrb.builtin.generator.common.task_input import (
    project_dir_input, app_name_input, module_name_input, entity_name_input,
    column_name_input, column_type_input
)
from zrb.builtin.generator.common.helper import validate_existing_project_dir
from zrb.builtin.group import project_add_group
from zrb.task.task import Task
from zrb.task.decorator import python_task
from zrb.runner import runner
from zrb.helper import util

import asyncio
import os

CURRENT_DIR = os.path.dirname(__file__)

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='validate',
    inputs=[
        project_dir_input, app_name_input, module_name_input, entity_name_input
    ],
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_existing_project_dir(project_dir)
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    entity_name = kwargs.get('entity_name')
    snake_module_name = util.to_snake_case(module_name)
    snake_entity_name = util.to_snake_case(entity_name)
    automation_dir = os.path.join(
        project_dir, '_automate', util.to_snake_case(app_name)
    )
    if not os.path.exists(automation_dir):
        raise Exception(
            f'Automation directory does not exist: {automation_dir}'
        )
    app_dir = os.path.join(
        project_dir, 'src', f'{util.to_kebab_case(app_name)}'
    )
    if not os.path.exists(app_dir):
        raise Exception(f'App directory does not exist: {app_dir}')
    module_path = os.path.join(app_dir, 'src', 'module', snake_module_name)
    if not os.path.exists(module_path):
        raise Exception(f'Module directory does not exist: {module_path}')
    entity_path = os.path.join(module_path, 'entity', snake_entity_name)
    if not os.path.exists(entity_path):
        raise Exception(f'Entity directory does not exist: {entity_path}')


@python_task(
    name='fastapp-field',
    group=project_add_group,
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
        entity_name_input,
        column_name_input,
        column_type_input,
    ],
    upstreams=[
        validate,
    ],
    runner=runner
)
async def add_fastapp_field(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    entity_name = kwargs.get('entity_name')
    column_name = kwargs.get('column_name')
    column_type = kwargs.get('column_name')
    kebab_app_name = util.to_kebab_case(app_name)
    kebab_module_name = util.to_kebab_case(module_name)
    snake_module_name = util.to_snake_case(module_name)
    kebab_entity_name = util.to_kebab_case(entity_name)
    snake_entity_name = util.to_snake_case(entity_name)
    pascal_entity_name = util.to_pascal_case(entity_name)
    snake_column_name = util.to_snake_case(column_name)
    kebab_column_name = util.to_kebab_case(column_name)
    capitalized_human_readable_column_name = util.to_capitalized_human_readable( # noqa
        column_name
    )
    await asyncio.gather(
        asyncio.create_task(add_column_to_test(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name, snake_column_name, column_type
        )),
        asyncio.create_task(add_column_to_schema(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name, pascal_entity_name, snake_column_name,
            column_type
        )),
        asyncio.create_task(add_column_to_repo(
            task, project_dir, kebab_app_name, snake_module_name,
            snake_entity_name, pascal_entity_name, snake_column_name,
            column_type
        )),
        asyncio.create_task(add_column_to_list_page(
            task, project_dir, kebab_app_name, kebab_module_name,
            kebab_entity_name, snake_column_name,
            capitalized_human_readable_column_name
        )),
        asyncio.create_task(add_column_to_detail_page(
            task, project_dir, kebab_app_name, kebab_module_name,
            kebab_entity_name, kebab_column_name, snake_column_name,
            capitalized_human_readable_column_name
        )),
        asyncio.create_task(add_column_to_delete_page(
            task, project_dir, kebab_app_name, kebab_module_name,
            kebab_entity_name, kebab_column_name, snake_column_name,
            capitalized_human_readable_column_name
        )),
        asyncio.create_task(add_column_to_update_page(
            task, project_dir, kebab_app_name, kebab_module_name,
            kebab_entity_name, kebab_column_name, snake_column_name,
            capitalized_human_readable_column_name
        )),
        asyncio.create_task(add_column_to_insert_page(
            task, project_dir, kebab_app_name, kebab_module_name,
            kebab_entity_name, kebab_column_name, snake_column_name,
            capitalized_human_readable_column_name
        )),
    )
    task.print_out('Success')



from zrb.task.task import Task
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_input.str_input import StrInput
import os


def test_task_with_no_runner():
    task = Task(
        name='simple-no-error',
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result


def test_task_with_predefined_runner():
    def _run(*args, **kwargs) -> str:
        return 'hello'
    task = Task(
        name='hello',
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'hello'


def test_task_with_skip_execution():
    def _run(*args, **kwargs) -> str:
        return 'hello'
    task = Task(
        name='hello',
        run=_run,
        retry=0,
        skip_execution=True
    )
    function = task.to_function()
    result = function()
    assert result is None


def test_task_with_input():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        favorite_drink = kwargs['favorite_drink']
        return f'hello {name}, your favorite drink is {favorite_drink}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name'),
            StrInput(name='favorite-drink')
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function(name='Dumbledore', favorite_drink='Elixir')
    assert result == 'hello Dumbledore, your favorite drink is Elixir'


def test_task_with_default_input():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        favorite_drink = kwargs['favorite_drink']
        return f'hello {name}, your favorite drink is {favorite_drink}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(name='favorite-drink', default='Elixir')
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'hello Nicholas Flamel, your favorite drink is Elixir'


def test_task_with_templated_input_without_kwarg():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        alias = kwargs['alias']
        return f'hello {name}, aka {alias}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(name='alias', default='{{input.name}}')
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'hello Nicholas Flamel, aka Nicholas Flamel'


def test_task_with_templated_input_and_partial_kwarg():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        alias = kwargs['alias']
        return f'hello {name}, aka {alias}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(name='alias', default='{{input.name}}')
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function(name='Alchemist')
    assert result == 'hello Alchemist, aka Alchemist'


def test_task_with_templated_input():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        alias = kwargs['alias']
        return f'hello {name}, aka {alias}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name', default='Nicholas Flamel'),
            StrInput(name='alias', default='{{input.name}}')
        ],
        run=_run,
        retry=0
    )
    # Name and alias provided
    function = task.to_function()
    result = function(name='Nicholas Flamel', alias='Alchemist')
    assert result == 'hello Nicholas Flamel, aka Alchemist'


def test_task_with_templated_env():
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('ZRB_TEST_TASK_ENV_2')
    task = Task(
        name='templated-env',
        envs=[
            Env(name='ZRB_TEST_TASK_ENV_1', default='Elixir'),
            Env(
                name='ZRB_TEST_TASK_ENV_2',
                default='{{env.ZRB_TEST_TASK_ENV_1}} of immortality'
            )
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    if 'ZRB_TEST_TASK_ENV_1' in os.environ:
        del os.environ['ZRB_TEST_TASK_ENV_1']
    if 'ZRB_TEST_TASK_ENV_2' in os.environ:
        del os.environ['ZRB_TEST_TASK_ENV_2']
    result = function()
    assert result == 'Elixir of immortality'


def test_task_with_env_file():
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('COLOR')
    env_file = os.path.join(os.path.dirname(__file__), 'test_task.env')
    task = Task(
        name='env-file',
        env_files=[EnvFile(env_file=env_file)],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'blue'


def test_task_with_env_file_and_prefix():
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('COLOR')
    env_file = os.path.join(os.path.dirname(__file__), 'test_task.env')
    task = Task(
        name='env-file-prefixed',
        env_files=[
            EnvFile(
                env_file=env_file, prefix='ZRB_TEST_TASK_WITH_ENV_AND_PREFIX'
            )
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    if 'ZRB_TEST_TASK_WITH_ENV_AND_PREFIX_COLOR' in os.environ:
        del os.environ['ZRB_TEST_TASK_WITH_ENV_AND_PREFIX_COLOR']
    os.environ['ZRB_TEST_TASK_WITH_ENV_AND_PREFIX_COLOR'] = 'red'
    result = function()
    assert result == 'red'
    del os.environ['ZRB_TEST_TASK_WITH_ENV_AND_PREFIX_COLOR']


def test_task_env_with_none_as_os_name():
    '''
    When env exist, it should override env_file
    '''
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('COLOR')
    env_file = os.path.join(os.path.dirname(__file__), 'test_task.env')
    task = Task(
        name='env-file-prefixed',
        env_files=[
            EnvFile(env_file=env_file, prefix='ZRB_TEST_TASK_OS_NAME_NONE')
        ],
        envs=[Env(name='COLOR', os_name=None, default='green')],
        run=_run,
        retry=0
    )
    function = task.to_function()
    if 'ZRB_TEST_TASK_OS_NAME_NONE_COLOR' in os.environ:
        del os.environ['ZRB_TEST_TASK_OS_NAME_NONE_COLOR']
    if 'COLOR' in os.environ:
        del os.environ['COLOR']
    os.environ['ZRB_TEST_TASK_OS_NAME_NONE_COLOR'] = 'red'
    os.environ['COLOR'] = 'cyan'
    result = function()
    assert result == 'cyan'
    del os.environ['ZRB_TEST_TASK_OS_NAME_NONE_COLOR']
    del os.environ['COLOR']


def test_task_env_with_empty_string_as_os_name():
    '''
    When env exist, it should override env_file,
    If env has empty os_name, the default value should always be used.
    '''
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('COLOR')
    env_file = os.path.join(os.path.dirname(__file__), 'test_task.env')
    task = Task(
        name='env-file-prefixed',
        env_files=[
            EnvFile(env_file=env_file, prefix='ZRB_TEST_TASK_OS_NAME_EMPTY')
        ],
        envs=[Env(name='COLOR', os_name='', default='green')],
        run=_run,
        retry=0
    )
    function = task.to_function()
    if 'ZRB_TEST_TASK_OS_NAME_EMPTY_COLOR' in os.environ:
        del os.environ['ZRB_TEST_TASK_OS_NAME_EMPTY_COLOR']
    if 'COLOR' in os.environ:
        del os.environ['COLOR']
    os.environ['ZRB_TEST_TASK_OS_NAME_EMPTY_COLOR'] = 'red'
    os.environ['COLOR'] = 'cyan'
    result = function()
    assert result == 'green'
    del os.environ['ZRB_TEST_TASK_OS_NAME_EMPTY_COLOR']
    del os.environ['COLOR']


def test_task_redeclared_env():
    '''
    When env exist, it should override env_file,
    If two env has the same name, the later should override the first one
    '''
    def _run(*args, **kwargs) -> str:
        task: Task = kwargs.get('_task')
        env_map = task.get_env_map()
        return env_map.get('COLOR')
    env_file = os.path.join(os.path.dirname(__file__), 'test_task.env')
    task = Task(
        name='env-file-prefixed',
        env_files=[EnvFile(env_file=env_file, prefix='ZRB_TEST_TASK')],
        envs=[
            Env(name='COLOR', os_name='', default='green'),
            Env(name='COLOR', os_name='', default='yellow')
        ],
        run=_run,
        retry=0
    )
    function = task.to_function()
    result = function()
    assert result == 'yellow'



from fastapi.responses import HTMLResponse
from config import app_src_dir
import os

index_html_path = os.path.join(app_src_dir, 'frontend', 'build', 'index.html')
with open(index_html_path, "r") as f:
    index_html_content = f.read()

frontend_index_response = HTMLResponse(
    content=index_html_content, status_code=200
)



from typeguard import typechecked
from zrb.runner import runner
from zrb.config.config import init_scripts, should_load_builtin
from zrb.helper.loader.load_module import load_module
from zrb.helper.log import logger
from zrb.helper.accessories.color import colored

import click
import os
import sys


@typechecked
def create_cli() -> click.Group:
    cli = click.Group(name='zrb', help='Super framework for your super app.')
    # load from ZRB_INIT_SCRIPTS environment
    for init_script in init_scripts:
        logger.info(colored(
            f'Load modules from {init_script}', attrs=['dark']
        ))
        load_module(script_path=init_script)
    # Load default tasks
    if should_load_builtin:
        logger.info(colored('Load builtins', attrs=['dark']))
        from zrb import builtin
        assert builtin
    # Load zrb_init.py
    project_dir = os.getenv('ZRB_PROJECT_DIR', os.getcwd())
    _load_zrb_init(project_dir)
    # Serve all tasks registered to runner
    logger.info(colored('Serve CLI', attrs=['dark']))
    cli = runner.serve(cli)
    return cli


@typechecked
def _load_zrb_init(project_dir: str):
    project_script = os.path.join(project_dir, 'zrb_init.py')
    if not os.path.isfile(project_script):
        return
    sys.path.append(project_dir)
    python_path = _get_new_python_path(project_dir)
    logger.info(colored(f'Set PYTHONPATH to {python_path}', attrs=['dark']))
    os.environ['PYTHONPATH'] = python_path
    logger.info(colored(f'Load modules from {project_script}', attrs=['dark']))
    load_module(script_path=project_script)


@typechecked
def _get_new_python_path(project_dir: str) -> str:
    current_python_path = os.getenv('PYTHONPATH')
    if current_python_path is None or current_python_path == '':
        return project_dir
    return current_python_path + ':' + project_dir



from typeguard import typechecked
import libcst as cst


@typechecked
class AddKeyValuePairTransformer(cst.CSTTransformer):
    def __init__(
        self, dict_name: str,
        key_node: cst.BaseExpression,
        value_node: cst.BaseExpression
    ):
        self.dict_name = dict_name
        self.key_node = key_node
        self.value_node = value_node
        super().__init__()

    def leave_Assign(
        self, original_node: cst.Assign, updated_node: cst.Assign
    ) -> cst.Assign:
        if (
            isinstance(updated_node.value, cst.Dict)
            and isinstance(updated_node.targets[0].target, cst.Name)
            and updated_node.targets[0].target.value == self.dict_name
        ):
            for element in updated_node.value.elements:
                key = element.key
                if (
                    isinstance(key, cst.BaseExpression)
                    and key.value == self.key_node
                ):
                    return updated_node
            new_elements = updated_node.value.elements + (
                cst.DictElement(self.key_node, self.value_node),
            )
            new_value = cst.Dict(new_elements)
            return updated_node.with_changes(value=new_value)
        return updated_node


def add_key_value_to_dict(
    code: str, dict_name: str, key: str, value: str
) -> str:
    module = cst.parse_module(code)
    key_node = cst.parse_expression(key)
    value_node = cst.parse_expression(value)
    transformed_module = module.visit(
        AddKeyValuePairTransformer(dict_name, key_node, value_node)
    )
    return transformed_module.code



from module.log.component.repo.activity_repo import (
    activity_repo
)
from module.log.entity.activity.model import (
    ActivityModel
)

activity_model: ActivityModel = ActivityModel(
    activity_repo
)



from typing import Any
from zrb.builtin.generator.fastapp_module.helper import (
    create_microservice_config, register_module, register_migration,
    create_app_config, append_all_enabled_env, append_all_disabled_env,
    append_src_template_env, append_deployment_template_env
)
from zrb.builtin.generator.common.helper import validate_existing_project_dir
from zrb.builtin.generator.common.task_input import (
    project_dir_input, app_name_input, module_name_input
)
from zrb.builtin.group import project_add_group
from zrb.task.task import Task
from zrb.task.decorator import python_task
from zrb.task.resource_maker import ResourceMaker
from zrb.runner import runner
from zrb.helper import util

import asyncio
import os

CURRENT_DIR = os.path.dirname(__file__)

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='validate',
    inputs=[project_dir_input, app_name_input, module_name_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_existing_project_dir(project_dir)
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    snake_module_name = util.to_snake_case(module_name)
    automation_dir = os.path.join(
        project_dir, '_automate', util.to_snake_case(app_name)
    )
    if not os.path.exists(automation_dir):
        raise Exception(
            f'Automation directory does not exist: {automation_dir}'
        )
    app_dir = os.path.join(
        project_dir, 'src', f'{util.to_kebab_case(app_name)}'
    )
    if not os.path.exists(app_dir):
        raise Exception(f'App directory does not exist: {app_dir}')
    module_path = os.path.join(app_dir, 'src', 'module', snake_module_name)
    if os.path.exists(module_path):
        raise Exception(f'Module directory already exists: {module_path}')


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        app_name_input,
        module_name_input,
    ],
    upstreams=[validate],
    replacements={
        'zrbAppName': '{{input.app_name}}',
        'zrbModuleName': '{{input.module_name}}',
    },
    template_path=os.path.join(CURRENT_DIR, 'template'),
    destination_path='{{ input.project_dir }}',
    excludes=[
        '*/__pycache__',
    ]
)


@python_task(
    name='fastapp-module',
    group=project_add_group,
    upstreams=[copy_resource],
    runner=runner
)
async def add_fastapp_module(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    kebab_app_name = util.to_kebab_case(app_name)
    snake_app_name = util.to_snake_case(app_name)
    kebab_module_name = util.to_kebab_case(module_name)
    snake_module_name = util.to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    await asyncio.gather(
        asyncio.create_task(create_microservice_config(
            task, project_dir, kebab_app_name, snake_app_name,
            kebab_module_name, snake_module_name, upper_snake_module_name
        )),
        asyncio.create_task(register_module(
            task, project_dir, kebab_app_name, snake_module_name
        )),
        asyncio.create_task(register_migration(
            task, project_dir, kebab_app_name, snake_module_name
        )),
        asyncio.create_task(create_app_config(
            task, project_dir, kebab_app_name, snake_module_name,
            upper_snake_module_name
        )),
        asyncio.create_task(append_all_enabled_env(
            task, project_dir, kebab_app_name, upper_snake_module_name
        )),
        asyncio.create_task(append_all_disabled_env(
            task, project_dir, kebab_app_name, upper_snake_module_name
        )),
        asyncio.create_task(append_src_template_env(
            task, project_dir, kebab_app_name, upper_snake_module_name
        )),
        asyncio.create_task(append_deployment_template_env(
            task, project_dir, kebab_app_name, upper_snake_module_name
        )),
    )
    task.print_out('Success')



from typing import Any, Mapping
from abc import ABC, abstractmethod
from core.error import HTTPAPIException
from module.auth.schema.token import AccessTokenData
from datetime import datetime, timedelta
from jose import jwt

import jsons


class AccessTokenUtil(ABC):

    @abstractmethod
    def encode(self, token_data: AccessTokenData) -> str:
        pass

    @abstractmethod
    def decode(
        self, token_str: str, parse_expired_token: bool = False
    ) -> AccessTokenData:
        pass


class JWTAccessTokenUtil(AccessTokenUtil):

    def __init__(self, secret_key: str, algorithm: str = 'HS256'):
        self.secret_key = secret_key
        self.algorithm = algorithm

    def encode(self, data: AccessTokenData) -> str:
        expire_time = datetime.utcnow() + timedelta(
            seconds=data.expire_seconds
        )
        sub = jsons.dumps(data.dict())
        data_dict = {'sub': sub, 'exp': expire_time}
        encoded_jwt = jwt.encode(
            data_dict, self.secret_key, algorithm=self.algorithm
        )
        return encoded_jwt

    def decode(
        self, token: str, parse_expired_token: bool = False
    ) -> AccessTokenData:
        try:
            decoded_data = jwt.decode(
                token,
                self.secret_key,
                algorithms=[self.algorithm],
                options=self._get_decode_options(parse_expired_token)
            )
            sub = jsons.loads(decoded_data['sub'])
            token_data = AccessTokenData.parse_obj(sub)
            if not parse_expired_token:
                expire_time = decoded_data['exp']
                token_data.expire_seconds = ((
                    datetime.fromtimestamp(expire_time) - datetime.utcnow()
                ).total_seconds())
                if token_data.expire_seconds < 0:
                    raise HTTPAPIException(422, 'Expired token')
            return token_data
        except jwt.JWTError:
            raise HTTPAPIException(422, 'Invalid token')

    def _get_decode_options(
        self, parse_expired_token: bool
    ) -> Mapping[str, Any]:
        if parse_expired_token:
            return {'verify_exp': False}
        return {}



from typing import Iterable, Optional
from typeguard import typechecked
from zrb.helper.accessories.color import colored

import datetime
import random
import re
import sys


@typechecked
class Advertisement():
    def __init__(self, content: str, time_pattern: str = '.*'):
        self.time_pattern = time_pattern
        self.content = content

    def show(self):
        print(colored(self.content, attrs=['dark']), file=sys.stderr)


@typechecked
def get_advertisement(
    advertisements: Iterable[Advertisement]
) -> Optional[Advertisement]:
    now = datetime.datetime.now().isoformat()
    candidates = [
        adv
        for adv in advertisements
        if re.match(adv.time_pattern, now)
    ]
    if len(candidates) == 0:
        return None
    return random.choice(candidates)



from typing import Iterable, Optional
from typeguard import typechecked
from termcolor import colored as term_colored, COLORS
import random


@typechecked
def get_random_color() -> str:
    colors = [
        'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', 'light_green',
        'light_yellow', 'light_blue', 'light_magenta', 'light_cyan',
    ]
    return random.choice(colors)


@typechecked
def is_valid_color(color: str) -> bool:
    return color in COLORS


@typechecked
def colored(
    text: str,
    color: Optional[str] = None,
    on_color: Optional[str] = None,
    attrs: Optional[Iterable[str]] = None
) -> str:
    return term_colored(text, color, on_color, attrs)



from sqlalchemy import create_engine, Engine
from sqlalchemy import event
from config import app_db_connection, app_db_engine_show_log
from component.log import logger
import time


@event.listens_for(Engine, "before_cursor_execute")
def before_cursor_execute(
    conn, cursor, statement, parameters, context, executemany
):
    context._query_start_time = time.time()
    one_line_statement = ' '.join(statement.split('\n'))
    logger.info(
        f'💽 [start query] SQL:`{one_line_statement}`, parameters {parameters}'
    )


@event.listens_for(Engine, "after_cursor_execute")
def after_cursor_execute(
    conn, cursor, statement,  parameters, context, executemany
):
    total = time.time() - context._query_start_time
    one_line_statement = ' '.join(statement.split('\n'))
    logger.info(' '.join([
        f'💽 [query completed] Elapsed time: {total} ms,',
        f'SQL: `{one_line_statement}`,',
        f'parameters {parameters}'
    ]))


engine: Engine = create_engine(app_db_connection, echo=app_db_engine_show_log)



from typing import Any, Optional
from core.messagebus.messagebus import (
    Publisher,  MessageSerializer, must_get_message_serializer
)
from core.messagebus.rabbitmq.admin import (
    RMQAdmin, must_get_rmq_admin
)
from pydantic import BaseModel

import aiormq
import asyncio
import logging


class RMQPublisher(Publisher):
    def __init__(
        self,
        logger: logging.Logger,
        connection_string: str,
        serializer: Optional[MessageSerializer] = None,
        rmq_admin: Optional[RMQAdmin] = None,
        retry: int = 5,
        retry_interval: int = 5,
        identifier='rmq-publisher'
    ):
        self.logger = logger
        self.rmq_admin = must_get_rmq_admin(
            logger=logger,
            rmq_admin=rmq_admin,
            connection_string=connection_string
        )
        self.connection_string = connection_string
        self.connection: Optional[aiormq.Connection] = None
        self.serializer = must_get_message_serializer(serializer)
        self.retry = retry
        self.retry_interval = retry_interval
        self.identifier = identifier

    async def publish(self, event_name: str, message: Any):
        await self.rmq_admin.create_events([event_name])
        queue_name = self.rmq_admin.get_queue_name(event_name)
        exchange_name = self.rmq_admin.get_exchange_name(event_name)
        if isinstance(message, BaseModel):
            message = message.dict()
        for attempt in range(self.retry):
            try:
                await self._connect()
                self.logger.info(f'🐰 [{self.identifier}] Get channel')
                self.logger.info(
                    f'🐰 [{self.identifier}] Publish to "{queue_name}": ' +
                    f'{message}'
                )
                await self.channel.basic_publish(
                    body=self.serializer.encode(event_name, message),
                    exchange=exchange_name,
                    routing_key=queue_name if exchange_name == '' else '',
                )
                return
            except (asyncio.CancelledError, GeneratorExit, Exception) as e:
                self.logger.error(
                    f'🐰 [{self.identifier}] Failed to publish message: {e}'
                )
                await self._disconnect()
                await asyncio.sleep(self.retry_interval)
                continue
        self.logger.error(
            f'🐰 [{self.identifier}] Failed to publish message after ' +
            f'{self.retry} attempts'
        )
        raise RuntimeError('Failed to publish message after retrying')

    async def _connect(self):
        try:
            connection_created = False
            if self.connection is None or self.connection.is_closed:
                self.logger.info(
                    f'🐰 [{self.identifier}] Create publisher connection'
                )
                self.connection = await aiormq.connect(self.connection_string)
                self.logger.info(
                    f'🐰 [{self.identifier}] Publisher connection created'
                )
                connection_created = True
            if (
                connection_created or
                self.channel is None or
                self.channel.is_closed
            ):
                self.logger.info(
                    f'🐰 [{self.identifier}] Get publisher channel'
                )
                self.channel = await self.connection.channel()
                self.logger.info(
                    f'🐰 [{self.identifier}] publisher channel created'
                )
        except (asyncio.CancelledError, GeneratorExit, Exception):
            self.logger.error(f'🐰 [{self.identifier}]', exc_info=True)
            raise Exception('Cannot connect')

    async def _disconnect(self):
        try:
            if self.channel is not None and not self.channel.is_closed:
                self.logger.info(
                    f'🐰 [{self.identifier}] Close publisher channel'
                )
                await self.channel.close()
                self.logger.info(
                    f'🐰 [{self.identifier}] Publisher channel closed'
                )
        except (asyncio.CancelledError, GeneratorExit, Exception):
            self.logger.error(f'🐰 [{self.identifier}]', exc_info=True)
        try:
            if self.connection is not None and not self.connection.is_closed:
                self.logger.info(
                    f'🐰 [{self.identifier}] Close publisher connection'
                )
                await self.connection.close()
                self.logger.info(
                    f'🐰 [{self.identifier}] Publisher connection closed'
                )
        except (asyncio.CancelledError, GeneratorExit, Exception):
            self.logger.error(f'🐰 [{self.identifier}]', exc_info=True)
        self.connection = None
        self.channel = None



from typing import Any, Callable, List, Mapping, TypeVar
from abc import ABC, abstractmethod

TRPCHandler = Callable[..., Any]
TMessage = TypeVar('TMessage', bound='Message')
TResult = TypeVar('TResult', bound='Result')


class Caller(ABC):
    @abstractmethod
    async def call(self, rpc_name: str, *args: Any, **kwargs: Any) -> Any:
        pass


class Server(ABC):
    @abstractmethod
    def register(self, rpc_name: str) -> Callable[[TRPCHandler], Any]:
        pass

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass


class Message():
    def __init__(
        self,
        reply_event: str,
        args: List[Any],
        kwargs: Mapping[str, Any]
    ):
        self.reply_event = reply_event
        self.args = args
        self.kwargs = kwargs

    def to_dict(self):
        return {
            'args': self.args,
            'kwargs': self.kwargs,
            'reply_event': self.reply_event
        }

    @classmethod
    def from_dict(cls, dictionary: Mapping[str, Any]) -> TMessage:
        return cls(
            reply_event=dictionary.get('reply_event', ''),
            args=dictionary.get('args', []),
            kwargs=dictionary.get('kwargs', {})
        )


class Result():
    def __init__(
        self,
        result: Any = None,
        error: str = ''
    ):
        self.result = result
        self.error = error

    def to_dict(self):
        return {
            'result': self.result,
            'error': self.error
        }

    @classmethod
    def from_dict(cls, dictionary: Mapping[str, Any]) -> TResult:
        return cls(
            result=dictionary.get('result', None),
            error=dictionary.get('error', '')
        )



from zrb.task_group.group import Group

project_group = Group(
    name='project', description='Project management'
)
project_add_group = Group(
    name='add', description='Add resources to project', parent=project_group
)
env_group = Group(
    name='env', description='Environment variable management'
)
ubuntu_group = Group(
    name='ubuntu', description='Ubuntu related commands'
)
ubuntu_install_group = Group(
    name='install', description='Install things on ubuntu', parent=ubuntu_group
)
explain_group = Group(
    name='explain', description='Explain things'
)
base64_group = Group(
    name='base64', description='Base64 operations'
)
md5_group = Group(
    name='md5', description='MD5 operations'
)
dev_tool_group = Group(
    name='devtool', description='Developer tools management'
)
dev_tool_install_group = Group(
    name='install', description='Install developer tools',
    parent=dev_tool_group
)
git_group = Group(
    name='git', description='Git related commands'
)



from typing import Any, Callable, Iterable, List, Mapping, Optional, Union
from typeguard import typechecked
from zrb.task.any_task import AnyTask
from zrb.task.base_task_composite import (
    AttemptTracker, FinishTracker, Renderer, TaskModelWithPrinterAndTracker
)
from zrb.advertisement import advertisements
from zrb.task_group.group import Group
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_input.any_input import AnyInput
from zrb.task_input.constant import RESERVED_INPUT_NAMES
from zrb.helper.accessories.color import colored
from zrb.helper.advertisement import get_advertisement
from zrb.helper.string.double_quote import double_quote
from zrb.helper.string.conversion import to_variable_name
from zrb.helper.map.conversion import to_str as map_to_str
from zrb.config.config import show_advertisement

import asyncio
import copy
import inspect
import os
import sys

MULTILINE_INDENT = ' ' * 8


@typechecked
class BaseTask(
    FinishTracker, AttemptTracker, Renderer, TaskModelWithPrinterAndTracker,
    AnyTask
):
    '''
    Base class for all tasks.
    Every task definition should be extended from this class.
    '''
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        description: str = '',
        inputs: List[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        retry: int = 2,
        retry_interval: float = 1,
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: float = 0,
        run: Optional[Callable[..., Any]] = None,
        skip_execution: Union[bool, str, Callable[..., bool]] = False
    ):
        # init properties
        retry_interval = retry_interval if retry_interval >= 0 else 0
        checking_interval = checking_interval if checking_interval > 0 else 0.1
        retry = retry if retry >= 0 else 0
        # init parent classes
        FinishTracker.__init__(self)
        Renderer.__init__(self)
        AttemptTracker.__init__(self, retry=retry)
        TaskModelWithPrinterAndTracker.__init__(
            self,
            name=name,
            group=group,
            description=description,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            retry=retry,
            retry_interval=retry_interval,
            upstreams=upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            run=run,
            skip_execution=skip_execution,
        )
        # init private properties
        self._is_keyval_set = False  # Flag
        self._all_inputs: Optional[List[AnyInput]] = None
        self._is_check_triggered: bool = False
        self._is_ready: bool = False
        self._is_execution_triggered: bool = False
        self._is_execution_started: bool = False
        self._args: List[Any] = []
        self._kwargs: Mapping[str, Any] = {}
        self._allow_add_upstream: bool = True

    def get_all_inputs(self) -> Iterable[AnyInput]:
        ''''
        Getting all inputs of this task and all its upstream, non-duplicated.
        '''
        if self._all_inputs is not None:
            return self._all_inputs
        self._allow_add_upstream = False
        self._all_inputs: List[AnyInput] = []
        existing_input_names: Mapping[str, bool] = {}
        # Add task inputs
        for task_input in self._inputs:
            if task_input.get_name() in existing_input_names:
                continue
            self._all_inputs.append(task_input)
            existing_input_names[task_input.get_name()] = True
        # Add upstream inputs
        for upstream in self._upstreams:
            upstream_inputs = upstream.get_all_inputs()
            for upstream_input in upstream_inputs:
                if upstream_input.get_name() in existing_input_names:
                    continue
                self._all_inputs.append(upstream_input)
                existing_input_names[upstream_input.get_name()] = True
        return self._all_inputs

    def to_function(
        self, env_prefix: str = '', raise_error: bool = True
    ) -> Callable[..., Any]:
        '''
        Return a function representing the current task.
        '''
        def function(*args: Any, **kwargs: Any) -> Any:
            self.log_info('Copy task')
            self_cp = copy.deepcopy(self)
            return asyncio.run(self_cp._run_and_check_all(
                env_prefix=env_prefix,
                raise_error=raise_error,
                args=args,
                kwargs=kwargs
            ))
        return function

    def add_upstreams(self, *upstreams: AnyTask):
        if not self._allow_add_upstream:
            raise Exception(f'Cannot add upstreams on `{self._name}`')
        self._upstreams += upstreams

    def inject_env_map(
        self, env_map: Mapping[str, str], override: bool = False
    ):
        '''
        Set new values for current task's env map
        '''
        for key, val in env_map.items():
            if override or key not in self.get_env_map():
                self._set_env_map(key, val)

    async def run(self, *args: Any, **kwargs: Any) -> Any:
        '''
        Do task execution
        Please override this method.
        '''
        if self._run_function is not None:
            if inspect.iscoroutinefunction(self._run_function):
                return await self._run_function(*args, **kwargs)
            return self._run_function(*args, **kwargs)
        return True

    async def check(self) -> bool:
        '''
        Return true when task is considered completed.
        By default, this will wait the task execution to be completed.
        You can override this method.
        '''
        return await self._is_done()

    def _show_done_info(self):
        complete_name = self._get_complete_name()
        elapsed_time = self._get_elapsed_time()
        message = '\n'.join([
            f'{complete_name} completed in {elapsed_time} seconds',
        ])
        self.print_out_dark(message)
        self._play_bell()

    def _get_multiline_repr(self, text: str) -> str:
        lines_repr: Iterable[str] = []
        lines = text.split('\n')
        if len(lines) == 1:
            return lines[0]
        for index, line in enumerate(lines):
            line_number_repr = str(index + 1).rjust(4, '0')
            lines_repr.append(f'{MULTILINE_INDENT}{line_number_repr} | {line}')
        return '\n' + '\n'.join(lines_repr)

    async def _set_local_keyval(
        self, kwargs: Mapping[str, Any], env_prefix: str = ''
    ):
        if self._is_keyval_set:
            return True
        self._is_keyval_set = True
        self.log_info('Set input map')
        for task_input in self.get_all_inputs():
            input_name = self._get_normalized_input_key(task_input.get_name())
            input_value = self.render_any(
                kwargs.get(input_name, task_input.get_default())
            )
            self._set_input_map(input_name, input_value)
        self.log_debug(
            'Input map:\n' + map_to_str(self.get_input_map(), item_prefix='  ')
        )
        self.log_info('Merging task envs, task env files, and native envs')
        for env_name, env in self._get_all_envs().items():
            env_value = env.get(env_prefix)
            if env.renderable:
                env_value = self.render_any(env_value)
            self._set_env_map(env_name, env_value)
        self.log_debug(
            'Env map:\n' + map_to_str(self.get_env_map(), item_prefix='  ')
        )

    def _get_all_envs(self) -> Mapping[str, Env]:
        all_envs: Mapping[str, Env] = {}
        for env_name in os.environ:
            all_envs[env_name] = Env(
                name=env_name, os_name=env_name, renderable=False
            )
        for env_file in self._env_files:
            for env in env_file.get_envs():
                all_envs[env.name] = env
        for env in self._envs:
            all_envs[env.name] = env
        return all_envs

    def _get_normalized_input_key(self, key: str) -> str:
        if key in RESERVED_INPUT_NAMES:
            return key
        return to_variable_name(key)

    async def _run_and_check_all(
        self,
        env_prefix: str,
        raise_error: bool,
        args: Iterable[Any],
        kwargs: Mapping[str, Any]
    ):
        try:
            self._start_timer()
            self.log_info('Set input and env map')
            await self._set_keyval(kwargs=kwargs, env_prefix=env_prefix)
            self.log_info('Set run kwargs')
            input_map = self.get_input_map()
            # make sure args and kwargs['_args'] are the same
            self.log_info('Set run args')
            new_args = copy.deepcopy(args)
            if len(args) == 0 and '_args' in kwargs:
                new_args = kwargs['_args']
            input_map['_args'] = new_args
            # inject self as input_map['_task']
            input_map['_task'] = self
            self._args = new_args
            self._kwargs = input_map
            # run the task
            coroutines = [
                asyncio.create_task(self._loop_check(show_info=True)),
                asyncio.create_task(self._run_all(*new_args, **input_map))
            ]
            results = await asyncio.gather(*coroutines)
            result = results[-1]
            self._print_result(result)
            return result
        except Exception as e:
            self.log_critical(f'{e}')
            if raise_error:
                raise
        finally:
            self._play_bell()

    def _print_result(self, result: Any):
        '''
        Print result to stdout so that it can be processed further.
        e.g.: echo $(zrb explain solid) > solid-principle.txt

        You need to override this method
        if you want to show the result differently.
        '''
        if result is None:
            return
        print(result)

    async def _loop_check(self, show_info: bool = False) -> bool:
        self.log_info('Start readiness checking')
        while not await self._cached_check():
            self.log_debug('Task is not ready')
            await asyncio.sleep(self._checking_interval)
        self._end_timer()
        self.log_info('State: ready')
        if show_info:
            if show_advertisement:
                selected_advertisement = get_advertisement(advertisements)
                selected_advertisement.show()
            self._show_done_info()
            self._show_run_command()
        return True

    def _show_run_command(self):
        params: List[str] = [double_quote(arg) for arg in self._args]
        for task_input in self.get_all_inputs():
            key = task_input.get_name()
            kwarg_key = self._get_normalized_input_key(key)
            quoted_value = double_quote(str(self._kwargs[kwarg_key]))
            params.append(f'--{key} {quoted_value}')
        run_cmd = self._get_complete_name()
        run_cmd_with_param = run_cmd
        if len(params) > 0:
            param_str = ' '.join(params)
            run_cmd_with_param += ' ' + param_str
        colored_run_cmd = colored(f'{run_cmd_with_param}', color='yellow')
        colored_label = colored('To run again: ', attrs=['dark'])
        print(colored(f'{colored_label}{colored_run_cmd}'), file=sys.stderr)

    async def _cached_check(self) -> bool:
        if self._is_check_triggered:
            self.log_debug('Waiting readiness flag to be set')
            while not self._is_ready:
                await asyncio.sleep(0.1)
            return True
        self._is_check_triggered = True
        check_result = await self._check()
        if check_result:
            self._is_ready = True
            self.log_debug('Set readiness flag to True')
        return check_result

    async def _check(self) -> bool:
        '''
        Check current task readiness.
        - If self.checkers is defined,
          this will return True once every self.checkers is completed
        - Otherwise, this will return check method's return value.
        '''
        if len(self._checkers) == 0:
            return await self.check()
        self.log_debug('Waiting execution to be started')
        while not self._is_execution_started:
            # Don't start checking before the execution itself has been started
            await asyncio.sleep(0.1)
        check_coroutines: Iterable[asyncio.Task] = []
        for checker_task in self._checkers:
            check_coroutines.append(
                asyncio.create_task(checker_task._run_all())
            )
        await asyncio.gather(*check_coroutines)
        return True

    async def _run_all(self, *args: Any, **kwargs: Any) -> Any:
        await self._mark_awaited()
        coroutines: Iterable[asyncio.Task] = []
        # Add upstream tasks to processes
        self._allow_add_upstream = False
        for upstream_task in self._upstreams:
            coroutines.append(asyncio.create_task(
                upstream_task._run_all(**kwargs)
            ))
        # Add current task to processes
        coroutines.append(self._cached_run(*args, **kwargs))
        # Wait everything to complete
        results = await asyncio.gather(*coroutines)
        return results[-1]

    async def _cached_run(self, *args: Any, **kwargs: Any) -> Any:
        if self._is_execution_triggered:
            self.log_debug('Task has been triggered')
            return
        self.log_info('State: triggered')
        self._is_execution_triggered = True
        self.log_info('State: waiting')
        # get upstream checker
        upstream_check_processes: Iterable[asyncio.Task] = []
        self._allow_add_upstream = False
        for upstream_task in self._upstreams:
            upstream_check_processes.append(asyncio.create_task(
                upstream_task._loop_check()
            ))
        # wait all upstream checkers to complete
        await asyncio.gather(*upstream_check_processes)
        # mark execution as started, so that checkers can start checking
        self._is_execution_started = True
        local_kwargs = dict(kwargs)
        local_kwargs['_task'] = self
        if await self._check_skip_execution(*args, **local_kwargs):
            self.log_info(
                f'Skip execution because config: {self._skip_execution}'
            )
            self.log_info('State: stopped')
            await self._mark_done()
            return None
        # start running task
        result: Any
        while self._should_attempt():
            try:
                self.log_debug(
                    f'Started with args: {args} and kwargs: {local_kwargs}'
                )
                self.log_info('State: started')
                result = await self.run(*args, **local_kwargs)
                break
            except Exception:
                self.log_info('State: failed')
                if self._is_last_attempt():
                    raise
                attempt = self._get_attempt()
                self.log_error(f'Encounter error on attempt {attempt}')
                self._increase_attempt()
                await asyncio.sleep(self._retry_interval)
                self.log_info('State: retry')
        await self._mark_done()
        self.log_info('State: stopped')
        return result

    async def _check_skip_execution(self, *args: Any, **kwargs: Any) -> bool:
        if callable(self._skip_execution):
            if inspect.iscoroutinefunction(self._skip_execution):
                return await self._skip_execution(*args, **kwargs)
            return self._skip_execution(*args, **kwargs)
        return self.render_bool(self._skip_execution)

    async def _set_keyval(self, kwargs: Mapping[str, Any], env_prefix: str):
        # if input is not in input_map, add default values
        for task_input in self.get_all_inputs():
            key = self._get_normalized_input_key(task_input.get_name())
            if key in kwargs:
                continue
            kwargs[key] = task_input.get_default()
        # set current task local keyval
        await self._set_local_keyval(kwargs=kwargs, env_prefix=env_prefix)
        # get new_kwargs for upstream and checkers
        new_kwargs = copy.deepcopy(kwargs)
        new_kwargs.update(self.get_input_map())
        upstream_coroutines = []
        # set uplstreams keyval
        self._allow_add_upstream = False
        for upstream_task in self._upstreams:
            upstream_coroutines.append(asyncio.create_task(
                upstream_task._set_keyval(
                    kwargs=new_kwargs, env_prefix=env_prefix
                )
            ))
        # set checker keyval
        local_env_map = self.get_env_map()
        checker_coroutines = []
        for checker_task in self._checkers:
            checker_task._inputs += self._inputs
            checker_task.inject_env_map(local_env_map, override=True)
            checker_coroutines.append(asyncio.create_task(
                checker_task._set_keyval(
                    kwargs=new_kwargs, env_prefix=env_prefix
                )
            ))
        # wait for checker and upstreams
        coroutines = checker_coroutines + upstream_coroutines
        await asyncio.gather(*coroutines)

    def __repr__(self) -> str:
        return f'<BaseTask name={self._name}>'



from fastapi import FastAPI
from logging import Logger
from core.messagebus import Publisher
from core.rpc import Caller
from module.auth.core import Authorizer


def register_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register API for "snake_zrb_module_name"')



from typing import Any, Callable, Iterable, Mapping, Optional, Union
from typeguard import typechecked
from zrb.task.base_task import BaseTask
from zrb.task.any_task import AnyTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput
from zrb.helper.file.copy_tree import copy_tree
from zrb.helper.util import (
    to_camel_case, to_pascal_case, to_kebab_case, to_snake_case,
    to_human_readable, to_capitalized_human_readable
)

Replacement = Mapping[str, str]
ReplacementMutator = Callable[
    [AnyTask, Replacement],
    Replacement
]


@typechecked
class ResourceMaker(BaseTask):

    def __init__(
        self,
        name: str,
        template_path: str,
        destination_path: str,
        replacements: Replacement = {},
        replacement_mutator: Optional[ReplacementMutator] = None,
        excludes: Iterable[str] = [],
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: Iterable[AnyTask] = [],
        skip_execution: Union[bool, str, Callable[..., bool]] = False,
        skip_parsing: Optional[Iterable[str]] = None
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=[],
            checking_interval=0.1,
            retry=0,
            retry_interval=0,
            skip_execution=skip_execution
        )
        self._template_path = template_path
        self._destination_path = destination_path
        self._excludes = excludes
        self._replacements = replacements
        self._replacement_mutator = replacement_mutator
        self._set_skip_parsing(skip_parsing)

    def _set_skip_parsing(self, skip_parsing: Optional[Iterable[str]]):
        if skip_parsing is not None:
            self._skip_parsing: Iterable[str] = skip_parsing
            return
        self._skip_parsing: Iterable[str] = [
            '*.mp3', '*.pdf', '*.exe', '*.dll', '*.bin', '*.iso', '*.png',
            '*.jpg', '*.gif', '*.ico'
        ]

    def to_function(
        self, env_prefix: str = '', raise_error: bool = True
    ) -> Callable[..., bool]:
        return super().to_function(env_prefix, raise_error)

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        # render parameters
        template_path = self.render_str(self._template_path)
        destination_path = self.render_str(self._destination_path)
        # render excludes
        self.log_debug(f'Render excludes: {self._excludes}')
        excludes = [self.render_str(exclude) for exclude in self._excludes]
        self.log_debug(f'Rendered excludes: {excludes}')
        # render replacements
        self.log_debug(f'Render replacements: {self._replacements}')
        rendered_replacements: Mapping[str, str] = {
            old: self.render_str(new)
            for old, new in self._replacements.items()
        }
        self.log_debug(f'Rendered replacements: {rendered_replacements}')
        if self._replacement_mutator is not None:
            self.log_debug('Apply replacement mutator')
            rendered_replacements = self._replacement_mutator(
                self, rendered_replacements
            )
        self.log_debug(
            f'Apply default replacement mutator: {rendered_replacements}'
        )
        rendered_replacements = self._default_mutate_replacements(
            rendered_replacements
        )
        self.log_debug(f'Final replacement: {rendered_replacements}')
        self.print_out_dark(f'Template: {template_path}')
        self.print_out_dark(f'Destination: {destination_path}')
        self.print_out_dark(f'Replacements: {rendered_replacements}')
        self.print_out_dark(f'Excludes: {excludes}')
        self.print_out_dark(f'Skip parsing: {self._skip_parsing}')
        await copy_tree(
            src=template_path,
            dst=destination_path,
            replacements=rendered_replacements,
            excludes=excludes,
            skip_parsing=self._skip_parsing
        )
        return True

    def _default_mutate_replacements(
        self, rendered_replacements: Mapping[str, str]
    ) -> Mapping[str, str]:
        transformations: Mapping[str, Callable[[str], str]] = {
            'Pascal': to_pascal_case,
            'kebab': to_kebab_case,
            'camel': to_camel_case,
            'snake': to_snake_case,
            'human readable': to_human_readable,
            'Human readable': to_capitalized_human_readable,
        }
        keys = list(rendered_replacements.keys())
        for key in keys:
            value = rendered_replacements[key]
            for prefix, transform in transformations.items():
                prefixed_key = transform(prefix + ' ' + key)
                if prefixed_key in rendered_replacements:
                    continue
                transformed_value = transform(value)
                rendered_replacements[prefixed_key] = transformed_value
        return rendered_replacements

    def __repr__(self) -> str:
        return f'<ResourceMaker name={self._name}>'



from zrb.helper.util import (
    coalesce, coalesce_str, to_camel_case, to_pascal_case, to_kebab_case,
    to_snake_case, to_human_readable, to_boolean
)
import datetime
import os
import platform
import time

DEFAULT_RENDER_DATA = {
    'datetime': datetime,
    'os': os,
    'platform': platform,
    'time': time,
    'util': {
        'coalesce': coalesce,
        'coalesce_str': coalesce_str,
        'to_camel_case': to_camel_case,
        'to_pascal_case': to_pascal_case,
        'to_kebab_case': to_kebab_case,
        'to_snake_case': to_snake_case,
        'to_human_readable': to_human_readable,
        'to_boolean': to_boolean,
    }
}



from zrb.task.cmd_task import CmdTask
from zrb.task_input.str_input import StrInput
import pathlib
import os


def test_cmd_task():
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello'
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == 'hello'


def test_cmd_task_with_error():
    cmd_task = CmdTask(
        name='simple-error',
        cmd='forbidden command'
    )
    function = cmd_task.to_function()
    is_error: bool = False
    try:
        function()
    except Exception:
        is_error = True
    assert is_error


def test_cmd_task_with_multiline_command():
    cmd_task = CmdTask(
        name='multiline-no-error',
        cmd=[
            'echo hello',
            'echo hello again',
        ]
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == '\n'.join(['hello', 'hello again'])


def test_cmd_task_with_function_returning_str_command():
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd=lambda *args, **kwargs: 'echo hello',
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == 'hello'


def test_cmd_task_with_function_returning_str_list_command():
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd=lambda *args, **kwargs: '\n'.join([
            'echo hello', 'echo hello again'
        ])
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == '\n'.join(['hello', 'hello again'])


def test_cmd_task_with_cmd_path():
    dir_path = pathlib.Path(__file__).parent.absolute()
    cmd_task = CmdTask(
        name='cmd-path-no-error',
        cmd_path=os.path.join(dir_path, 'templates', 'hello-world.sh')
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == 'Hello, World!'


def test_cmd_task_with_multiline_cmd_path():
    dir_path = pathlib.Path(__file__).parent.absolute()
    cmd_task = CmdTask(
        name='cmd-path-no-error',
        cmd_path=[
            os.path.join(dir_path, 'templates', 'hello-heaven.sh'),
            os.path.join(dir_path, 'templates', 'hello-world.sh')
        ]
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == '\n'.join(['Hello, Heaven!', 'Hello, World!'])


def test_cmd_task_with_function_returning_str_cmd_path():
    dir_path = pathlib.Path(__file__).parent.absolute()
    cmd_path = os.path.join(dir_path, 'templates', 'hello-world.sh')
    cmd_task = CmdTask(
        name='cmd-path-no-error',
        cmd_path=lambda *args, **kwargs: cmd_path
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == 'Hello, World!'


def test_cmd_task_with_function_returning_str_list_cmd_path():
    dir_path = pathlib.Path(__file__).parent.absolute()
    cmd_path = [
        os.path.join(dir_path, 'templates', 'hello-heaven.sh'),
        os.path.join(dir_path, 'templates', 'hello-world.sh')
    ]
    cmd_task = CmdTask(
        name='cmd-path-no-error',
        cmd_path=lambda *args, **kwargs: cmd_path
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == '\n'.join(['Hello, Heaven!', 'Hello, World!'])


def test_cmd_task_with_upstream_with_no_error():
    upstream_task = CmdTask(
        name='upstream-no-error',
        cmd='echo upstream'
    )
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello',
        upstreams=[upstream_task]
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == 'hello'


def test_cmd_task_with_upstream_with_error():
    upstream_task = CmdTask(
        name='upstream-error',
        cmd='forbidden command'
    )
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello',
        upstreams=[upstream_task]
    )
    function = cmd_task.to_function()
    is_error: bool = False
    try:
        function()
    except Exception:
        is_error = True
    assert is_error


def test_cmd_task_with_upstream_containing_inputs():
    upstream = CmdTask(
        name='ask-name',
        inputs=[
            StrInput(name='name')
        ]
    )
    cmd_task = CmdTask(
        name='hello',
        cmd='echo hello {{input.name}}',
        upstreams=[upstream]
    )
    function = cmd_task.to_function()
    result = function(name='Dumbledore')
    assert result.output == 'hello Dumbledore'


def test_cmd_task_with_diamond_upstream():
    upstream_0 = CmdTask(
        name='upstream-no-error',
        cmd='echo upstream 0'
    )
    upstream_1 = CmdTask(
        name='upstream-no-error',
        cmd='echo upstream 1',
        upstreams=[upstream_0]
    )
    upstream_2 = CmdTask(
        name='upstream-no-error',
        cmd='echo upstream 2',
        upstreams=[upstream_0]
    )
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello',
        upstreams=[upstream_1, upstream_2]
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == 'hello'


def test_cmd_task_with_checker_with_no_error():
    checker = CmdTask(
        name='check',
        cmd='echo checked'
    )
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello',
        checkers=[checker]
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == 'hello'


def test_cmd_task_with_checker_with_error():
    checker = CmdTask(
        name='check',
        cmd='forbidden command'
    )
    cmd_task = CmdTask(
        name='simple-no-error',
        cmd='echo hello',
        checkers=[checker]
    )
    function = cmd_task.to_function()
    is_error: bool = False
    try:
        function()
    except Exception:
        is_error = True
    assert is_error


def test_cmd_task_with_long_output():
    cmd_task = CmdTask(
        name='long-output',
        cmd=[
            'echo 1',
            'echo 2',
            'echo 3',
            'echo 4',
            'echo 5',
        ],
        max_output_line=3,
    )
    function = cmd_task.to_function()
    result = function()
    assert result.output == '\n'.join(['3', '4', '5'])



from sqlalchemy import Column, Integer, Table, ForeignKey
from module.snake_zrb_module_name.component.base import Base

'''
You should put any "many-to-many" table declarations in this file.
For example:
```
# Many-to-many table to relate users and permissions
user_permission = Table(
    "user_permission",
    Base.metadata,
    Column("user_id", Integer, ForeignKey("users.id")),
    Column("permission_id", Integer, ForeignKey("permissions.id")),
)
```

Remove the following assertions:
'''
assert Column
assert Integer
assert Table
assert ForeignKey
assert Base



from component.log import logger
from component.db_connection import engine
from module.snake_zrb_module_name.entity.snake_zrb_entity_name.repo import (
    PascalZrbEntityNameRepo, PascalZrbEntityNameDBRepo
)

snake_zrb_entity_name_repo: PascalZrbEntityNameRepo = PascalZrbEntityNameDBRepo(
    logger=logger, engine=engine
)



from typeguard import typechecked
from typing import Mapping


@typechecked
def parse_replacement(text: str, replacement: Mapping[str, str]):
    new_text = text
    for old, new in replacement.items():
        new_text = new_text.replace(old, new)
    return new_text



from .build_project_images import build_project_images
from .deploy_project import deploy_project
from .push_project_images import push_project_images
from .remove_project_containers import remove_project_containers
from .destroy_project import destroy_project
from .start_project_containers import start_project_containers
from .start_project import start_project
from .stop_project_containers import stop_project_containers

assert build_project_images
assert deploy_project
assert push_project_images
assert remove_project_containers
assert destroy_project
assert start_project_containers
assert start_project
assert stop_project_containers
